package xyz.lychee.lagfixer.nms.v1_15_R1;

import com.google.common.collect.ImmutableSet;
import java.lang.reflect.Field;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Stream;
import net.minecraft.server.v1_15_R1.CriterionTriggers;
import net.minecraft.server.v1_15_R1.EntityAgeable;
import net.minecraft.server.v1_15_R1.EntityAnimal;
import net.minecraft.server.v1_15_R1.EntityCreature;
import net.minecraft.server.v1_15_R1.EntityExperienceOrb;
import net.minecraft.server.v1_15_R1.EntityHuman;
import net.minecraft.server.v1_15_R1.EntityPlayer;
import net.minecraft.server.v1_15_R1.EntityTameableAnimal;
import net.minecraft.server.v1_15_R1.GameRules;
import net.minecraft.server.v1_15_R1.Item;
import net.minecraft.server.v1_15_R1.Items;
import net.minecraft.server.v1_15_R1.PathfinderGoal;
import net.minecraft.server.v1_15_R1.PathfinderGoalBreed;
import net.minecraft.server.v1_15_R1.PathfinderGoalSelector;
import net.minecraft.server.v1_15_R1.PathfinderGoalTempt;
import net.minecraft.server.v1_15_R1.PathfinderGoalWrapped;
import net.minecraft.server.v1_15_R1.PathfinderTargetCondition;
import net.minecraft.server.v1_15_R1.StatisticList;
import net.minecraft.server.v1_15_R1.World;
import net.minecraft.server.v1_15_R1.PathfinderGoal.Type;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.craftbukkit.v1_15_R1.entity.CraftCreature;
import org.bukkit.craftbukkit.v1_15_R1.event.CraftEventFactory;
import org.bukkit.entity.Chicken;
import org.bukkit.entity.Cow;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Horse;
import org.bukkit.entity.Parrot;
import org.bukkit.entity.Pig;
import org.bukkit.entity.Sheep;
import org.bukkit.entity.Villager;
import org.bukkit.event.entity.EntityBreedEvent;
import org.bukkit.event.entity.EntityTargetLivingEntityEvent;
import org.bukkit.event.entity.CreatureSpawnEvent.SpawnReason;
import org.bukkit.event.entity.EntityTargetEvent.TargetReason;
import xyz.lychee.lagfixer.modules.CustomAiModule;
import xyz.lychee.lagfixer.utils.ReflectionUtils;

public class CustomAi extends CustomAiModule.NMS {
   private final PathfinderGoal optimized = new PathfinderGoal() {
      public boolean a() {
         return false;
      }
   };
   private final boolean collides;
   private final boolean silent;
   private final boolean typeAenabled;
   private final boolean typeBenabled;
   private final boolean typeCenabled;
   private final Set<String> blacklistSet;
   private final Set<String> whitelistSet;
   private final PathfinderTargetCondition temptTargeting;
   private final PathfinderTargetCondition breedTargeting;
   private final double temptSpeed;
   private final double breedSpeed;
   private final int temptCooldown;
   private final boolean temptTriggerBothHands;
   private final boolean temptEvent;
   private final boolean temptTeleport;
   private final boolean breedEvent;
   private final boolean breedTeleport;
   private final Field pgwField;

   public CustomAi(CustomAiModule module) {
      super(module);
      ConfigurationSection section = this.getModule().getSection();
      this.collides = section.getBoolean("collides");
      this.silent = section.getBoolean("silent");
      this.typeAenabled = section.getBoolean("pathfinder.typeA.enabled");
      this.typeBenabled = section.getBoolean("pathfinder.typeB.enabled");
      this.typeCenabled = section.getBoolean("pathfinder.typeC.enabled");
      this.whitelistSet = ImmutableSet.copyOf(section.getStringList("pathfinder.typeA.whitelist"));
      this.blacklistSet = ImmutableSet.copyOf(section.getStringList("pathfinder.typeB.blacklist"));
      this.temptTargeting = this.targetingConditions(section, "animals.tempt").d().c();
      this.breedTargeting = this.targetingConditions(section, "animals.breed").c();
      this.temptSpeed = section.getDouble("animals.tempt.speed");
      this.breedSpeed = section.getDouble("animals.breed.speed");
      this.temptCooldown = section.getInt("animals.tempt.cooldown");
      this.temptTriggerBothHands = section.getBoolean("animals.tempt.trigger_two_hands");
      this.temptEvent = section.getBoolean("animals.tempt.event");
      this.temptTeleport = section.getBoolean("animals.tempt.teleport");
      this.breedEvent = section.getBoolean("animals.breed.event");
      this.breedTeleport = section.getBoolean("animals.breed.teleport");
      this.pgwField = ReflectionUtils.getPrivateField(PathfinderGoalSelector.class, Set.class);
   }

   private PathfinderTargetCondition targetingConditions(ConfigurationSection section, String key) {
      return (new PathfinderTargetCondition()).a(section.getDouble(key + ".range")).a().b();
   }

   public void setCustomAi(Entity ent, boolean init) {
      if (ent instanceof CraftCreature) {
         CraftCreature cc = (CraftCreature)ent;
         EntityCreature mob = cc.getHandle();
         if (cc.getHandle().targetSelector.c().noneMatch((pgw) -> {
            return pgw.h() == -1;
         })) {
            mob.collides = this.collides;
            mob.setSilent(this.silent);
            mob.targetSelector.a(-1, this.optimized);

            try {
               Set<PathfinderGoalWrapped> goals = (Set)this.pgwField.get(mob.goalSelector);
               if (this.typeAenabled) {
                  goals.removeIf((pgw) -> {
                     return !pgw.j().getClass().getName().contains("$") && this.whitelistSet.stream().noneMatch((str) -> {
                        return pgw.j().getClass().getSimpleName().contains(str);
                     });
                  });
               } else if (this.typeBenabled) {
                  goals.removeIf((pgw) -> {
                     return this.blacklistSet.stream().anyMatch((str) -> {
                        return pgw.j().getClass().getSimpleName().contains(str);
                     });
                  });
               } else if (this.typeCenabled) {
                  goals.clear();
               }

               if (mob instanceof EntityAnimal) {
                  goals.removeIf((pgw) -> {
                     return pgw.j().getClass().equals(PathfinderGoalBreed.class);
                  });
                  goals.add(new PathfinderGoalWrapped(0, new OptimizedBreedGoal((EntityAnimal)mob)));
               }

               Item item = ent instanceof Villager ? Items.EMERALD : (ent instanceof Horse ? Items.APPLE : (ent instanceof Pig ? Items.CARROT : (!(ent instanceof Cow) && !(ent instanceof Sheep) ? (!(ent instanceof Chicken) && !(ent instanceof Parrot) ? null : Items.WHEAT_SEEDS) : Items.WHEAT)));
               if (item != null) {
                  goals.removeIf((pgw) -> {
                     return pgw.j().getClass().equals(PathfinderGoalTempt.class);
                  });
                  goals.add(new PathfinderGoalWrapped(1, new OptimizedTemptGoal(mob, item)));
               }
            } catch (IllegalAccessException var7) {
               IllegalAccessException ex = var7;
               this.getModule().getPlugin().printError(ex);
            }
         }
      }

   }

   public class OptimizedBreedGoal extends PathfinderGoal {
      protected final EntityAnimal animal;
      protected EntityAnimal partner;

      public OptimizedBreedGoal(EntityAnimal entityanimal) {
         this.animal = entityanimal;
         this.a(EnumSet.of(Type.MOVE));
      }

      public boolean a() {
         if (CustomAi.this.breedEvent) {
            EntityTargetLivingEntityEvent event = CraftEventFactory.callEntityTargetLivingEvent(this.animal, this.partner, TargetReason.CUSTOM);
            if (event.isCancelled()) {
               return false;
            }
         }

         return this.animal.isInLove() && (this.partner = this.getFreePartner()) != null;
      }

      public boolean b() {
         return this.partner.isAlive() && this.partner.isInLove();
      }

      public void e() {
         if (CustomAi.this.breedTeleport) {
            this.animal.teleportAndSync(this.partner.locX(), this.partner.locY(), this.partner.locZ());
         } else {
            this.animal.getNavigation().a(this.partner, CustomAi.this.breedSpeed);
         }

         EntityAgeable child = this.animal.createChild(this.partner);
         if (child != null) {
            if (child instanceof EntityTameableAnimal && ((EntityTameableAnimal)child).isTamed()) {
               child.persistent = true;
            }

            EntityPlayer player = this.animal.getBreedCause();
            if (player == null && this.partner.getBreedCause() != null) {
               player = this.partner.getBreedCause();
            }

            EntityBreedEvent event = CraftEventFactory.callEntityBreedEvent(child, this.animal, this.partner, player, this.animal.breedItem, this.animal.getRandom().nextInt(7) + 1);
            if (event.isCancelled()) {
               return;
            }

            if (player != null) {
               player.a(StatisticList.ANIMALS_BRED);
               CriterionTriggers.o.a(player, this.animal, this.partner, child);
            }

            this.animal.setAgeRaw(6000);
            this.animal.resetLove();
            this.partner.setAgeRaw(6000);
            this.partner.resetLove();
            child.setAgeRaw(-24000);
            child.setPositionRotation(this.animal.locX(), this.animal.locY(), this.animal.locZ(), 0.0F, 0.0F);
            World world = this.animal.getWorld();
            world.addEntity(child, SpawnReason.BREEDING);
            world.broadcastEntityEffect(this.animal, (byte)18);
            int experience = event.getExperience();
            if (world.getGameRules().getBoolean(GameRules.DO_MOB_LOOT) && experience > 0) {
               world.addEntity(new EntityExperienceOrb(world, this.animal.locX(), this.animal.locY(), this.animal.locZ(), experience));
            }
         }

      }

      private EntityAnimal getFreePartner() {
         List<? extends EntityAnimal> nearbyEntities = this.animal.world.a(this.animal.getClass(), CustomAi.this.breedTargeting, this.animal, this.animal.getBoundingBox().g(8.0));
         EntityAnimal var10000;
         if (nearbyEntities.isEmpty()) {
            var10000 = null;
         } else {
            Stream var2 = nearbyEntities.stream();
            EntityAnimal var10001 = this.animal;
            Objects.requireNonNull(var10001);
            var2 = var2.filter(var10001::mate);
            var10001 = this.animal;
            Objects.requireNonNull(var10001);
            var10000 = (EntityAnimal)var2.min(Comparator.comparingDouble(var10001::h)).orElse((Object)null);
         }

         return var10000;
      }
   }

   public class OptimizedTemptGoal extends PathfinderGoal {
      private final EntityCreature mob;
      private final PathfinderTargetCondition targetingConditions;
      private final Item item;
      private int cooldown = 0;

      public OptimizedTemptGoal(EntityCreature mob, Item item) {
         this.mob = mob;
         this.item = item;
         this.a(EnumSet.of(Type.MOVE));
         this.targetingConditions = CustomAi.this.temptTargeting.a();
      }

      public boolean a() {
         return this.cooldown-- <= 0;
      }

      public void e() {
         EntityHuman player;
         label42: {
            this.cooldown = CustomAi.this.temptCooldown;
            player = this.mob.world.a(this.targetingConditions, this.mob);
            if (player != null) {
               if (CustomAi.this.temptTriggerBothHands) {
                  if (player.getItemInMainHand().getItem() == this.item || player.getItemInOffHand().getItem() == this.item) {
                     break label42;
                  }
               } else if (player.getItemInMainHand().getItem() == this.item) {
                  break label42;
               }
            }

            return;
         }

         if (CustomAi.this.temptEvent) {
            EntityTargetLivingEntityEvent event = CraftEventFactory.callEntityTargetLivingEvent(this.mob, player, TargetReason.TEMPT);
            if (event.isCancelled()) {
               return;
            }
         }

         if (this.mob.h(player) < 6.25 && !CustomAi.this.temptTeleport) {
            this.mob.getNavigation().o();
         } else if (CustomAi.this.temptTeleport) {
            this.mob.teleportAndSync(player.locX(), player.locY(), player.locZ());
         } else {
            this.mob.getNavigation().a(player, this.mob instanceof EntityAnimal ? CustomAi.this.temptSpeed : 0.35);
         }

      }
   }
}
